# -*- coding: utf-8 -*-
"""
Created on Thu Jul 11 14:02:55 2013

@author: Christian
"""


from PyQt4.QtCore import *
from PyQt4.QtGui import *

from schemaobject import SchemaObject
from processblock import ProcessBlock

from port import __PORT_TYPE_INPUT__



class CanevasSchema(SchemaObject): #, QGraphicsView

    # SIGNALS
    
    # Emitted when the focus has changed from one object to another in the canevas.
    # This allows, for exemple, to update the object properties window. The newly
    # selected object is passed as parameter.
    selectedSchemaObjectChanged = pyqtSignal(object)

    # Emitted when the logic of the schema has changed. This can occur for many reason:
    # a block has been added/deleted, a link has been added/deleted, etc. The argument is
    # a vector of the objects for which the logic has been altered. For example, when a
    # new link is established between two blocs, the logic of two SchemaObject has been alterred.
    schemaLogicChanged          = pyqtSignal(list) 

    def __init__(self, parent):
        super(CanevasSchema, self).__init__(parent)
    
        # Dict of ProcessBlock objects indexed by the block's names (string)
        self.mapBlock = {}

        self.lstLinks = []

        # True if presently creating a link between two blocks
        self.bCreatingLink = False
    
        # Link object being created
        self.link = None
    
        # Port type from the port initiating the link creation.
        self.initiatingPort = None
    
        # To be passed as parameter when the signal
        # schemaLogicChanged(...) is emmited.
        self.modifiedObjects = []
    
        self.setAcceptDrops(True)
        
        Pal = self.palette()
        # set black background
        Pal.setColor(QPalette.Background, Qt.white)
        self.setAutoFillBackground(True)
        self.setPalette(Pal)
    
        self.setFocusPolicy(Qt.StrongFocus)



    def addBlock(self, processBlock, strBlockName=""):
        
        # If no name proposed, we use the name of the block class,
        # which we take in lowercase to differentiate the instance (lowercase)
        # and the     std::map<string, shared_ptr<ProcessBlock> > m_mapBlock;
        if strBlockName == "":

            strBlockName = processBlock.getName()
            #transform(strBlockName.begin(), strBlockName.end(), strBlockName.begin(), ::tolower)
            strBlockName = strBlockName.lower()

    
        # If the proposed name already exist, we append an integer
        # requential number
        tmpStr = strBlockName
        i      = 1
        # If the key is already in use in m_mapBlock
        while tmpStr in self.mapBlock :
            tmpStr = strBlockName +  str(i)
            i += 1
    
        # Storing the block inside the caneva
        self.mapBlock[tmpStr] = processBlock
    
        
        processBlock.selectedSchemaObjectChanged.connect(self.emitSelectedSchemaObjectChanged)
        processBlock.updateEvent.connect(self.update)
    
    
        # Store the absolute position
        globalPos = processBlock.mapToGlobal(processBlock.pos())
    
        # Reparenting to make it being a child of the caneva.
        processBlock.setParent(self)
    
        # Restore the absolute position to cancel the shift occuring
        # when reparenting.
        processBlock.move(processBlock.mapFromGlobal(globalPos))
    
        # Make sure that it shows.
        self.mapBlock[tmpStr].show()
    
        self.modifiedObjects = []
        self.modifiedObjects.append(processBlock)
        #emitSchemaLogicChanged();
    




    def emitSelectedSchemaObjectChanged(self, objectSelected):
        self.selectedSchemaObjectChanged.emit(objectSelected)
        

    def emitSchemaLogicChanged(self):        
        self.schemaLogicChanged.emit(self.modifiedObjects)


    def finishingALink(self, receivingPort) :
        
        if receivingPort and self.initiatingPort :
            self.modifiedObjects = []
            self.modifiedObjects.append(receivingPort.parentWidget())
            self.modifiedObjects.append(self.initiatingPort.parentWidget())
            self.emitSchemaLogicChanged()
    
        self.link               = None
        self.initiatingPort     = []
        self.bCreatingLink      = False



    def creatingALink(self, link, initPort):
        self.link           = link
        self.initiatingPort = initPort
        self.bCreatingLink  = True


    def dragMoveEvent(self, event):
        if self.bCreatingLink:
            if not self.link is None :
                if self.initiatingPort.getPortType() & __PORT_TYPE_INPUT__ :
                    self.link.setP1(event.pos())
                else:
                    self.link.setP2(event.pos())
    
                #self.link.redefine()
    

    def dragEnterEvent(self, event):
        if self.bCreatingLink:
            event.acceptProposedAction()



    def isCreatingALinkself(self) : return self.bCreatingLink
    def getLinkInitPort(self)     : return self.initiatingPort

    # For walking though the schema.
    def getNbPreviousObject(self)   : return 0
    def getNbNextObject(self)       : return 0
    def previousObject(self, index) : return None
    def nextObject(self, index)     : return None


"""


class CanevasSchema : public SchemaObject
{
    Q_OBJECT
private:
    std::map<string, shared_ptr<ProcessBlock> > m_mapBlock;
    std::list<shared_ptr<Link> > lstLinks;

    // To be passed as parameter when the signal
    // schemaLogicChanged(...) is emmited.
    vector<SchemaObjectPtr> m_modifiedObjects;

    // True if presently creating a link between two blocks
    bool m_bCreatingLink;

    // Link object being created
    shared_ptr<Link> m_ptrLink;

    // Port type from the port initiating the link creation.
    PortPtr m_pInitiatingPort;

public:
    explicit CanevasSchema(QWidgetPtr parent);
    ~CanevasSchema();

    void        init() {};

    void addBlock(shared_ptr<ProcessBlock> r_ptrBlock, string r_strBlockName = "");



    void finishingALink(PortPtr r_ptrReceivingBlock);
    void creatingALink(shared_ptr<Link> r_ptrLink, PortPtr r_pInitPort);



private slots:
    virtual void formatObjectCode() {};



private slots:
    void emitSchemaLogicChanged();

"""
