# -*- coding: utf-8 -*-
"""
Created on Thu Jul 11 17:03:25 2013

@author: Christian
"""

from PyQt4.QtCore import *
from PyQt4.QtGui import *


from schemaobject import SchemaObject
from constant import PORT_SIZE
from link import ArrowLink

__PORT_TYPE_OUTPUT__  = 0x0000
__PORT_TYPE_INPUT__   = 0x0001



class Port(SchemaObject):

    # Is there currently a drag-and-drop in operation from a Port object
    inDragAndDrop = False

    updateEvent = pyqtSignal(QRegion)
    
    def __init__(self, strName, parent, portType):
        
        SchemaObject.__init__(self, parent)

        self.leftMouseButtonDown = False

        # Is a candidate (if the mouse button is release before the
        # cursor go out of reach)
        # to receive the drop from a drag and drop.
        self.isDoppingCandidate = False
    
        # Port link object
        self.link = None
    
        self.linkedPort = []
    
        # Name of the port
        self.strName = strName
    
        # Last recorded position of the mouse
        self.lastMousePos = None
    
        # Variables defining the shape of the port
        self.portPoly = QPolygon([QPoint(0, 0), QPoint(PORT_SIZE, PORT_SIZE/2), QPoint(0, PORT_SIZE)])
        
    
        self.portType = portType;

        if self.portType & __PORT_TYPE_INPUT__:
            self.linkingPoint = QPoint(0, PORT_SIZE/2);
        else:
            self.linkingPoint = QPoint(PORT_SIZE, PORT_SIZE/2);


        self.setAcceptDrops(True)
        self.setAutoFillBackground(True)
        self.setBackgroundRole(QPalette.Window)
        self.setFocusPolicy(Qt.StrongFocus);
        
        mask = QBitmap(parent.size())
        mask.clear()
        painter = QPainter()
        painter.begin(mask)
        painter.setPen(Qt.color1)
        painter.setBrush(Qt.color1)
        painter.drawConvexPolygon(self.portPoly)
        painter.end()
        self.setMask(mask)
    
    def setLink(self, link):
        self.link = link
        self.link.updateEvent.connect(self.updateEvent)
    
        
    def changeLinkP2(self, point):
        if self.link:
            self.link.setP2(point)

    
    
    def changeLinkP1(self, point):
        if self.link :
            self.link.setP1(point)

        
    
    
    
    
    
    def getLinkedPortName(self):
        if self.linkedPort:
            return self.linkedPort.getPortName()
        return ""
    
    def getLinkP2(self):
        return self.link.p2() if self.link else QPoint(0, 0)
    

    def getLinkP1(self):
        return self.link.p1()  if self.link else QPoint(0, 0)
    


    # For walking though the schema. For ports, these functions return arrow linked to them.
    def getNbPreviousObject(self) : return 1 if self.isLinked() and self.isInput()  else 0
    def getNbNextObject(self)     : return 1 if seld.isLinked() and self.isOutput() else 0
    def previousObject(self)      : return self.getLinkPtr() if self.getNbPreviousObject() else None
    def nextObject(self)          : return self.getLinkPtr() if self.getNbNextObject()     else None


    def getPortType(self)         : return self.portType
    def isInput(self)             : return self.portType and __PORT_TYPE_INPUT__
    def isOutput(self)            : return not self.isInput()
    def getPortName(self)         : return self.strName

    def getLink(self)             : return self.link
    def isLinked(self)            : return self.link != None
    def getLinkedPort(self)       : return self.linkedPort



    #private slots:
    #    virtual void formatObjectCode() {};



    
    def paintEvent(self, event) :
        Pal = self.palette()
        # set black background
        if self.hasFocus() or self.isDoppingCandidate :
            Pal.setColor(QPalette.Background, Qt.red)
        else :
            Pal.setColor(QPalette.Background, Qt.black)
        self.setPalette(Pal)
        QWidget.paintEvent(self, event)
    
    
    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton :
            self.leftMouseButtonDown = True
            self.lastMousePos        = event.globalPos()
    
    
    def mouseReleaseEvent(self, event) :
        if event.button() == Qt.LeftButton:
            Port.inDragAndDrop         = False
            self.leftMouseButtonDown   = False

    
    def mouseMoveEvent(self, event) :
        if self.leftMouseButtonDown :
            if not Port.inDragAndDrop :
                # TODO: Double call to parent might be dangerous. Might be better to use
                # events.
                holder = self.getHolder()
    
                self.setLink(ArrowLink(holder)) # the caneva is the parent
    
                if self.portType & __PORT_TYPE_INPUT__:
                    self.link.setP2(self.mapTo(holder, self.linkingPoint))
                else:
                    self.link.setP1(self.mapTo(holder, self.linkingPoint))
    
                holder.creatingALink(self.link, self)
    
                drag     = QDrag(self)
                mimeData =  QMimeData()
                mimeData.setText("")
    
                drag.setMimeData(mimeData)
    
                Port.inDragAndDrop = True
                drag.exec_()
                Port.inDragAndDrop = False
    
    
    
    
    def dragEnterEvent(self, event):
        if Port.inDragAndDrop:
            # TODO: Double call to parent might be dangerous. Might be better to use
            # events.
            holder          = self.getHolder()
            pInitiatingPort = holder.getLinkInitPort()
    
            # The initiating port and the receiving port for the connection must be
            # complementary (one input, one output)
            if (self.portType & __PORT_TYPE_INPUT__ ) != (pInitiatingPort.getPortType() & __PORT_TYPE_INPUT__) :
                link  = pInitiatingPort.getLink()
    
                self.isDoppingCandidate = True
                event.acceptProposedAction()
    
                if self.portType & __PORT_TYPE_INPUT__ :
                    link.setP2(self.mapTo(holder, self.linkingPoint))
                else:
                    link.setP1(self.mapTo(holder, self.linkingPoint))
    
                #link.redefine()
    
                self.update()

    
    
    def dragLeaveEvent(self, event):
        self.isDoppingCandidate = False
        self.update()
    
    
    
    def dropEvent(self, event):
    
        if Port.inDragAndDrop :
            # TODO: Double call to parent might be dangerous. Might be better to use
            # events.
            holder = self.getHolder()
    
            self.linkedPort = holder.getLinkInitPort()
            
            # The initiating port and the receiving port for the connection must be
            # complementary (one input, one output)
            if (self.portType & __PORT_TYPE_INPUT__) != (self.linkedPort.getPortType() & __PORT_TYPE_INPUT__) :
                self.linkedPort.linkedPort = self
                self.setLink(self.linkedPort.getLink())
    
                holder.finishingALink(self)
                event.acceptProposedAction()

    
    def moveEvent(self, event):
        if self.link:
            if self.portType & __PORT_TYPE_INPUT__:
                self.link.setP2(self.mapTo(self.getHolder(), self.linkingPoint))
            else :
                self.link.setP1(self.mapTo(self.getHolder(), self.linkingPoint))
    
            self.link.redefine()
    

    # Ports are entry port of an SchemaObject which is hold in a container. This
    # function return a smart pointer on that container.
    def getHolder(self) :
        return self.parentWidget().parentWidget()
