# -*- coding: utf-8 -*-
"""
Created on Thu Jul 11 15:10:41 2013

@author: Christian
"""


from PyQt4.QtCore import *
from PyQt4.QtGui import *

from scipy import arange

from schemaobject import SchemaObject
from sizepoint import SizePoint, SIZEPOINT_TOP, SIZEPOINT_BOTTOM, SIZEPOINT_LEFT, SIZEPOINT_RIGHT
from constant import PORT_SIZE
from port import Port, __PORT_TYPE_INPUT__,  __PORT_TYPE_OUTPUT__

class ProcessBlock(SchemaObject):

    ###########################################################################
    # SIGNALS
    ###########################################################################
    
    
    selectedSchemaObjectChanged = pyqtSignal(object)
    
    updateEvent = pyqtSignal(QRegion)
    
    # Emitted when this schema object has been moved such that other objects
    # can update accordingly.
    #schemaObjectMoved = pyqtSignal()




    def __init__(self, parent, strName):

        SchemaObject.__init__(self, parent)
       
        # FOR THE DRAG AND DROP OPERATION
        self.bDragging = False
        self.lastMousePos = None

        # FOR THE APPEARANCE
        self.lblBox = QLabel(self)
    
        # FOR RESIZING
        self.resizePointsTop            = SizePoint(self, SIZEPOINT_TOP)
        self.resizePointsBottom         = SizePoint(self, SIZEPOINT_BOTTOM)
        self.resizePointsLeft           = SizePoint(self, SIZEPOINT_LEFT)
        self.resizePointsRight          = SizePoint(self, SIZEPOINT_RIGHT)
        self.resizePointsTopRight       = SizePoint(self, SIZEPOINT_RIGHT | SIZEPOINT_TOP)
        self.resizePointsTopLeft        = SizePoint(self, SIZEPOINT_LEFT  | SIZEPOINT_TOP)
        self.resizePointsBottomRight    = SizePoint(self, SIZEPOINT_RIGHT | SIZEPOINT_BOTTOM)
        self.resizePointsBottomLeft     = SizePoint(self, SIZEPOINT_LEFT  | SIZEPOINT_BOTTOM)

        # Name of the block. It must be unique whithin a schema.
        self.strName = strName
    
        # List of input and output ports
        self.lstInputPort  = []
        self.lstOutputPort = []        
            
    
        self.setAcceptDrops(True)        
   
        self.setFocusPolicy(Qt.StrongFocus)

        palette = QPalette()
        palette.setColor(QPalette.Foreground, Qt.red)
        self.setPalette(palette);

        # Setting the block name
        self.lblBox.setText(self.strName)
        self.lblBox.setStyleSheet("border: 1px solid black")
        self.lblBox.setStyleSheet("background: white")
        self.lblBox.setFrameStyle(QFrame.Box)
        self.lblBox.setAlignment(Qt.AlignVCenter | Qt.AlignHCenter)


    def guessBestGeometry(self):
        
        metrics     = QFontMetrics(self.lblBox.font())
        pixelWidth  = max(metrics.boundingRect(self.lblBox.text()).width()*1.2,                20) + 4*PORT_SIZE
        pixelHeight = max(max(len(self.lstInputPort), len(self.lstOutputPort))*(PORT_SIZE+10), 20) + 2*PORT_SIZE
        
        self.setGeometry(self.geometry().x(), self.geometry().y(),
                         pixelWidth, pixelHeight)


    def setLastMousePos(self, pos) :
        self.lastMousePos = pos


    def addInputPort(self, strName):
        port = Port(strName, self, __PORT_TYPE_INPUT__ )
        self.lstInputPort.append(port)
        port.updateEvent.connect(self.updateEvent.emit)


    def addOutputPort(self, strName):
        port = Port(strName, self, __PORT_TYPE_OUTPUT__)
        self.lstOutputPort.append(port)
        port.updateEvent.connect(self.updateEvent.emit)

    
    
    
    def setResizeVisible(self, isVisible):
        self.resizePointsTop.setVisible(isVisible)
        self.resizePointsBottom.setVisible(isVisible)
        self.resizePointsLeft.setVisible(isVisible)
        self.resizePointsRight.setVisible(isVisible)
        self.resizePointsTopRight.setVisible(isVisible)
        self.resizePointsTopLeft.setVisible(isVisible)
        self.resizePointsBottomRight.setVisible(isVisible)
        self.resizePointsBottomLeft.setVisible(isVisible)
    
    def previousObject(self, iIndex):
        if iIndex >= self.getNbPreviousObject() :
            return None
    
        if not self.lstInputPort[iIndex].isLinked() :
            return None
    
        return self.lstInputPort[iIndex].getLinkedPort().parentWidget()
    
    
    def nextObject(self, iIndex):
        if iIndex >= self.getNbNextObject():
            return None
    
        if not self.lstOutputPort[iIndex].isLinked():
            return None
    
        return self.lstOutputPort[iIndex].getLinkedPort().parentWidgetPtr()

    
    def getName(self):
        return self.strName

    # For walking though the schema. For blocks, next and previous objects are other blocks.
    def getNbPreviousObject(self):
        return len(self.lstInputPort)
        
    def getNbNextObject(self) :      
        return len(self.lstOutputPort)
        
        
        
    ###########################################################################
    # EVENTS 
    ###########################################################################

    # Events overloading
    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton :
            self.bDragging = True
            self.lastMousePos = event.globalPos()


    def mouseReleaseEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.bDragging = False;
    

    def mouseMoveEvent(self, event):
        
        if self.bDragging:   # The block is being moved.
        
            self.move(self.pos() + event.globalPos() - self.lastMousePos)

            for port in self.lstInputPort:
                port.changeLinkP2(port.getLinkP2() + event.globalPos() - self.lastMousePos)
                
    
            for port in self.lstOutputPort:
                port.changeLinkP1(port.getLinkP1() + event.globalPos() - self.lastMousePos)
    
            self.lastMousePos = event.globalPos() 
            #self.schemaObjectMoved.emit()
    

    def focusInEvent(self, event):
        self.setFrameStyle(QFrame.Box)
        self.setResizeVisible(True)
        super(ProcessBlock, self).focusInEvent(event)

    
    def focusOutEvent(self, event):
        self.setFrameStyle(QFrame.NoFrame)
        self.setResizeVisible(False)
    
    
    
    def resizeEvent(self, event):
        self.resizePointsTop.move(        self.width()/2, 0)
        self.resizePointsBottom.move(     self.width()/2, self.height()-self.resizePointsBottom.height())
        self.resizePointsLeft.move(       0,         self.height()/2)
        self.resizePointsRight.move(      self.width()-self.resizePointsRight.width(),   self.height()/2)
        self.resizePointsTopRight.move(   self.width()-self.resizePointsTopRight.width(),   0);
        self.resizePointsTopLeft.move(    0,         0);
        self.resizePointsBottomRight.move(self.width()-self.resizePointsBottomRight.width(),   self.height()-self.resizePointsBottomRight.height());
        self.resizePointsBottomLeft.move( 0,         self.height()-self.resizePointsBottomLeft.height());
    
        self.lblBox.setGeometry(PORT_SIZE, PORT_SIZE, self.width()-2*PORT_SIZE, self.height()-2*PORT_SIZE);
    
        nbItem = len(self.lstInputPort)
        for port, i2 in zip(self.lstInputPort, arange(nbItem)+1):
            port.move(PORT_SIZE, self.size().height()*i2/(nbItem+1) - PORT_SIZE/2)

        nbItem = len(self.lstOutputPort)
        for port, i2 in zip(self.lstOutputPort, arange(nbItem)+1):
            port.move(self.size().width()-PORT_SIZE, self.size().height()*i2/(nbItem+1) - PORT_SIZE/2)
    
    
            
        
        
