# -*- coding: utf-8 -*-
"""
Created on Wed May 29 14:14:39 2013

@author: REVESTECH
"""


from abc import ABCMeta, abstractmethod 
from copy import deepcopy
import sys
import Pyro4
from PyBlockWork import Register, StagingRoom
import uuid
import serpent



class Block():
    __metaclass__ = ABCMeta
    
    """
     inputs  : Is a dictionnary of variableLabels:variableValues. Big objects
               should be pickled and only the file name should be passed because
               input objects are used in the object signature.
     outputs : A dictionnaly of output variableLabels:outputProperties:
               "parallel": for an object that which all elements can be proceed
                           independently.
               "atomic"  : for an object that which...
    """
    def __init__(self, inputs, outputs, verbose = True, ID = None):

        for inputKey in inputs:
            self.__dict__[inputKey] = inputs[inputKey]

        if ID is None:
            self.ID = uuid.uuid1()
        else:
            self.ID = ID

        self.verbose  = verbose
        self.outputs  = outputs

        # We keep this as it is used in buliding the signature of the block.
        self.inputs = inputs

        self.nextBlockIDs = []
        
        # Tells that this block is an entry point (the first block of the schema).
        # This is set to false before any call to the start() method of a subsequent
        # block, such that only the block started by the user has this value as True.
        #self.isEntryPoint = True
        
        # True if these is a process is avaiable. It should always be set to
        # True by the preceeding blocks, passing its available processing slot,
        # except for the first block of the schema which need to request it
        # directly from the manager.
        #self.processAvailable = False        

        sys.excepthook = Pyro4.util.excepthook    
        self.registerName = None
        self.stagingRoomName  = None


    def changeID(self):
        self.ID = uuid.uuid1()


    #def setSchedulerName(self, name):
    #    self.schedulerName = name

    def addNextBlockID(self, ID):
        self.nextBlockIDs.append(ID)


    """
     Should be implemented with the code to execute. In this code, input and
     output variables must be preceded by "self."
    """
    @abstractmethod
    def execute(self): 
        raise "Error: Pure virtual function"

    """
     Return the signature of the BWM.
    """
    def getSignature(self):
        signature = str(type(self)) + ":" + str(self.inputs) + ":" + str(self.outputs)
        if "fromPreviousOutput" in self.__dict__ :
            signature +=  ":" + str(self.fromPreviousOutput)
        return signature

    
    #def setEntryPoint(self, isEntryPoint):
    #    self.isEntryPoint = isEntryPoint



    #def getSchemaStructure(self):
    #
    #    return {"name":self.__class__.__name__,
    #            "inputs":self.inputs,
    #            "outputs":self.outputs,
    #            "nexts":[b.getSchemaStructure() for b in self.nextBlocks] }


    def print_(self, string):
        try:
            with Pyro4.Proxy("PYRONAME:" + self.pyroInfos["comHubName"]) as comHub:   
                comHub.writeMessage(self.workerID, string)  
        except Exception:
            print "Pyro traceback:"
            print "".join(Pyro4.util.getPyroTraceback())   
            raise




    def registerExec(self, func, *args, **kargs):
        if self.registerName is None:
            raise Exception("The block register name is set to None. Connection to the register not possible.")
        
        try:
            with Pyro4.Proxy("PYRONAME:" + self.registerName) as register:   
                #return func(register, *args, **kargs)  
            
                if func == Register.hasExecuted:
                    return register.hasExecuted(args[0])
                elif func == Register.getOutputDict:
                    return register.getOutputDict(args[0])  
                elif func == Register.registerExecutionStart:
                    register.registerExecutionStart(args[0])  
                elif func == Register.registerExecutionCompletion:
                    register.registerExecutionCompletion(args[0], args[1])            
            
        except Exception:
            print "Pyro traceback:"
            print "".join(Pyro4.util.getPyroTraceback()) 
            raise
        
    

    def stagingRoomExec(self, func, *args, **kargs):
        if self.stagingRoomName is None:
            raise Exception("The staging room name is set to None. Connection to the staging room not possible.")
        
        try:
            with Pyro4.Proxy("PYRONAME:" + self.stagingRoomName) as stagingRoom:   
                
                # __func__ is used to get around the fact that stagingRoom is not of type
                # StagingRoom but Proxy, which causes problems in python 
                # 2.6+ for unbound methods.
                #return func.__func__(stagingRoom, *args, **kargs)  

                if func == StagingRoom.registerBlockString:
                    stagingRoom.registerBlockString(args[0], args[1])
                elif func == StagingRoom.getBlockString:
                    return stagingRoom.getBlockString(args[0])
                   
        except Exception:
            print "Pyro traceback:"
            print "".join(Pyro4.util.getPyroTraceback()) 
            raise
        
        
    
    def stageBlock(self, ID):
        try:
            with Pyro4.Proxy("PYRONAME:" + self.stagingRoomName) as stagingRoom:   
                stagingRoom.addReadyBlock(ID)
        except Exception:
            print "Pyro traceback:"
            print "".join(Pyro4.util.getPyroTraceback())       
            raise
        
        
        
      

    """
     The function used to start the processing of the block. Normally, it should
     by a BlockWorker object.
    """
    def run(self, workerID, pyroInfos): 
        
        if self.registerName is None:
            raise ValueError("A register name must have been set to Block objects before they can be ran.")
        
        self.pyroInfos = pyroInfos        
        
        if not isinstance(pyroInfos, dict):
            raise TypeError("pyroInfos must be a dict object.")

        if "stagingRoomName" in pyroInfos:
            self.stagingRoomName = pyroInfos["stagingRoomName"]
            
        if self.stagingRoomName is None:
            raise NameError("pyrInfos must contain the stagingRoomName key if block.statingRoomName has been left to None.")                     
            
        for key in ["comHubName", "HMAC_KEY"]:                                                 
            if not key in pyroInfos:
                raise NameError("pyrInfos must contain the key " + key + ".")        
        
        self.workerID = workerID
        Pyro4.config.HMAC_KEY = pyroInfos["HMAC_KEY"]           
        
        #print "Nb Next blocks:", len(self.nextBlocks)
        #sys.stdout.flush()            
        

        # Some outputes are coming from the previous block.    
        if "fromPreviousOutput" in self.__dict__ :
            for key in self.fromPreviousOutput:
                self.__dict__[key] = self.fromPreviousOutput[key]
                self.inputs[key]   = self.fromPreviousOutput[key]        

        
        try:
            Pyro4.locateNS()
        except Pyro4.errors.PyroError, e:
            print 'Pyro error message: ' + str(e)
            raise
        except:
            raise
        
        #== self.scheduler = Pyro4.Proxy("PYRONAME:" + self.schedulerName)                     
        
              
 
        #try:
        
        # If this is true, this block is the entry point of the schema. We must
        # therefore tell the Scheduler about the schema structure for it to be
        # able to gives information necessary to display the schema.
        #            
        # PROBLEM: Setting the schema structure this way make the scheduler
        # unserializable because of too deep structure caused by chaining
        # of blocks.
        #if self.isEntryPoint :        
        #    self.scheduler.setSchemaStructure(self.getSchemaStructure())
                           
                   
                   
        # If self.processAvailable is true, the available slot has been passed
        # from a previous block so there is no need to ask it to the scheduler.
        # Else, it must be gathered from the scheduler.
        #== if not self.processAvailable :         
        #==     self.scheduler.acquireProcess()    
        #==     self.processAvailable = True
            
        if self.registerExec(Register.hasExecuted, self.getSignature()):                    
            self.print_(self.getSignature() + " has already been processed...")
            
            # Getting the output of the block as it may be necessary for
            # the execution of next blocks.
            outputDict = self.registerExec(Register.getOutputDict, self.getSignature())
                                       
            for key in outputDict:
                self.__dict__[key] = outputDict[key]
        else:
            # Indicating that the module has ben entered.
            self.print_(self.getSignature() + " started...")
           
            self.registerExec(Register.registerExecutionStart, self.getSignature())
                
            # Executing the module's code.
            self.execute()
            
            # Indicating that the module's code execution is terminated.
            outputDict = dict((k,self.__dict__[k]) for k in self.outputs.keys()  if k in self.__dict__)
            self.registerExec(Register.registerExecutionCompletion, self.getSignature(), outputDict)
            self.print_(self.getSignature() + " completed..."    )                    
            
        # We want to promote in depth execution rather than lateral one, to avoid
        # generating too much processes. Thus, we favor that each block gives the 
        # priority to its first next block. For that, we do not release the available 
        # process but pass it directly to the first next block.
                   

        # Running next blocks....
        for blockID in self.nextBlockIDs:
            
            blockString = self.stagingRoomExec(StagingRoom.getBlockString, blockID)
            block = Block.deserialize(blockString)
                        
            nextInputs = None
            # Verify if the block needs input from this block.
            for inputKey in block.inputs:
                                      
                # If this input needs to come from the output of this block
                if isinstance(block.__dict__[inputKey], fromPrevious):
                    
                    # verify if this block has a corresponding output.
                    if block.__dict__[inputKey].outputKey in self.outputs:
                        
                        # Is the output atomic or parallel
                        if self.outputs[block.__dict__[inputKey].outputKey] == "atomic":
                            if nextInputs is None : 
                                nextInputs = [{inputKey:self.__dict__[block.__dict__[inputKey].outputKey]}]
                            else:
                                for nextInput in nextInputs:
                                    nextInput[inputKey] = self.__dict__[block.__dict__[inputKey].outputKey]
                            
                        elif self.outputs[block.__dict__[inputKey].outputKey] == "parallel":
                            
                            # Every nextInput needs to be duplicated for every items of
                            # the parallel output.
                            temp = []
                            for item in self.__dict__[block.__dict__[inputKey].outputKey]:
                                if nextInputs is None : 
                                    temp.append({inputKey:item})
                                else:                                
                                    for nextInput in nextInputs:
                                        copyNI = deepcopy(nextInput)
                                        copyNI[inputKey] = item
                                        temp.append(copyNI)

                            nextInputs = temp
                        
                        else:
                            self.print_("Unknown output type.")                                   
                            raise NameError
                        
                    else: # there is no corresponding output
                        self.print_("No output/input correspondance: " + \
                                          block.__dict__[inputKey].outputKey +\
                                         " is not in available outputs " + str(self.outputs))   
                        raise NameError 
             

            if nextInputs is None :  
                #block.setEntryPoint(False)
                self.stageBlock(block.ID)                           
                   
            else :   
                for nextInput in nextInputs:                   
                    # A thread can only be started once so we need to copy
                    # it as many times as required. 
                    newblock = deepcopy(block)
                    newblock.fromPreviousOutput = nextInput                           
                    #newblock.setEntryPoint(False)
                    newblock.changeID()
                    self.stagingRoomExec(StagingRoom.registerBlockString, newblock.ID, newblock.serialize())
                    self.stageBlock(newblock.ID)                             



                

        # This error occurs when the scheduler has been closed. There is no
        # action to take as the wanted behavior in this case is for the current
        # process to end, which happends at the end of this function.
        #except Pyro4.errors.ConnectionClosedError:
        #    pass            
        #except Pyro4.errors.CommunicationError:
        #    pass            
        #except:
        #    self.print_("Error running : " + self.getSignature())
        #    if self.processAvailable:
        #        self.scheduler.releaseProcess()
        #    raise


    def serialize(self):       
        serpentString = serpent.dumps(self, indent=False, set_literals=False)
        
        # Removing header because it conflicts with Pyros serializing
        serpentString = "\n".join(serpentString.split('\n')[1:])
        return serpentString



        
    @staticmethod
    def deserialize(serializedString):

        # Putting back the header removed at serialization
        header = serpent.dumps(1, indent=True, set_literals=False).split('\n')[0]

        AST = serpent.loads(header + "\n" + serializedString)
        
        return Block.fromASTtoObject(AST)
                            



    @staticmethod
    def fromASTtoObject(AST):
        className = AST["__class__"]            
        
        inputs = AST["inputs"]
        for inputKey in inputs :
            if isinstance(inputs[inputKey], dict) and '__class__' in inputs[inputKey] :  
                if inputs[inputKey]['__class__'] == "fromPrevious" :
                    inputs[inputKey] = fromPrevious(inputs[inputKey]['outputKey'])
                    #inputs[inputKey] = "fromPrevious('" + inputs[inputKey]['outputKey'] + "')"                        
        
        
        keywords = {"inputs":inputs, "outputs":AST["outputs"],
                    "verbose":AST["verbose"], "ID":AST["ID"]}

        from PyBlockWork.blocks import *
        block = apply(locals()[className], [], keywords)  
        
        block.nextBlockIDs          = AST["nextBlockIDs"]
        block.registerName          = AST["registerName"]
        
        if "fromPreviousOutput" in AST:
            block.fromPreviousOutput    = AST["fromPreviousOutput"]

        return block

        

"""
 Class used to identified the BWM input variables as comming form a previous
 block.
"""
class fromPrevious:
    def __init__(self, outputKey):
        self.outputKey = outputKey
    def __repr__(self):
        return "fromPrevious(outputKey='" + str(self.outputKey) + "')"






