# -*- coding: utf-8 -*-
"""
Created on Wed Jun 19 13:42:13 2013

@author: oreichri
"""


import PyBlockWork
import sys


from spyndle.propagation import computeXCST, computeSPF
from spyndle.io import HarmonieReader, EDFReader
from spyndle.detector import SpindleDectectorRMS
from spyndle.EEG import getActiveElectrode
from spyndle.io import Session


"""
     Inputs  : "fileSIG" : String indicating the full path to a .sig file.
                          
     Outputs : "files"   : String indicating the full path to a .bdf file.
"""
class SigBdfConversionBWM(PyBlockWork.Block):

    def execute(self):   
        if self.verbose:        
            self.print_("Reading " + self.fileSIG + "...")
            
        readerSIG =  HarmonieReader(self.fileSIG)

        if self.verbose:             
            self.print_("Saving a copy in the BDF format...")
            
        readerSIG.saveAsEDF(self.fileSIG[:-3] + "bdf", "BDF", verbose=False)    
        
        # Setting outputs
        self.fileBDF = self.fileSIG[:-3] + "bdf"






"""
     Inputs  : "file"               : String indicating the full path to an EEG file.
               "readerClass"        : 
               "reformattingExpr"   :
                          
     Outputs : "fileReformatted"    : String indicating the full path to the reformated file.
"""
class ChannelReformattingBWM(PyBlockWork.Block):

    def execute(self):   
        if self.verbose:        
            self.print_("Reading " + self.file + "...")

        if hasattr(self, 'readerClass'):
            if not isinstance(self.readerClass, (str, unicode)) :
                raise TypeError("Expected type str or unicode, received type " + str(type(self.readerClass)) + ".")
            reader = apply(globals()[self.readerClass], [self.file])  
        else:
            reader =  EDFReader(self.file)              
            

        self.fileReformatted = self.file[:-4] + "_reformatted" + self.file[-4:]

        if self.verbose:             
            self.print_("Saving a reformated copy as " + self.fileReformatted  + "...")

        reader.reformatMontage(self.reformattingExpr, self.fileReformatted)








                               
"""
     Inputs  : "file" : String indicating the full path to a .sig file.
               "includePatterns" : List of channel patterns to be listed. 
                                   (e.g., ['Fp1', 'Fp2', 'F7')
               "excludePatterns" : List of channel patterns to be excluded from
                                   the listing (e.g., ['Fp1', 'Fp2', 'F7').
               "readerClass" : (optional) Specify the reader class to be used.
                                EDFReader is used by default.
                          
     Outputs : "files"   : String indicating the full path to EEG file.
"""
class GetChannelListBWM(PyBlockWork.Block):

    def execute(self): 
                
        # A reader class can be spedicied as en input but the EDFReader is used
        # by default.

        if hasattr(self, 'readerClass'):
            if not isinstance(self.readerClass, (str, unicode)) :
                raise TypeError("Expected type str or unicode, received type " + str(type(self.readerClass)) + ".")
            reader = apply(globals()[self.readerClass], [self.file])  
        else:
            reader =  EDFReader(self.file)              
            
            
        
        channelLabels = reader.getChannelLabels()

        OK = len(channelLabels)*[False]
        for ind, channel in enumerate(channelLabels):
            if len([include for include in self.includePatterns if include in channel]):
                OK[ind] = True
            if len([include for include in self.excludePatterns if include in channel]):
                OK[ind] = False                
                
        self.listChannels = [channel for ind, channel in enumerate(channelLabels) if OK[ind]]
        rejectedChannels = [channel for ind, channel in enumerate(channelLabels) if not OK[ind]]

        if self.verbose:
            self.print_("Selected channels:" + str(self.listChannels))
            self.print_("Rejected channels:" + str(rejectedChannels))
            sys.stdout.flush()




"""
     Inputs  : "file" : String indicating the full path to a .sig file.
               "listChannels" : List of channel on which we want to extract 
                                the spindles(e.g., ['Fp1-ref', 'Fp2-ref', 'F7-ref')
               "readerClass"  : (optional) Specify the reader class to be used.
                                EDFReader is used by default.
               "detectorClass": (optional) Specify the detector class to be used.
                                 SpindleDectectorRMS is used by default.
               "eventName"    : (optional) Specify the name of the spindle event.
                                 "SpindleRMS" is used by default.
               "detectionStages" : (optional) List of detection stages. 
                                   Default value : ["Sleep stage 2"]
               "detectionThreshold" : (optional) Detection threshold. 
               "computeSlopeFreq"   : (optional) Specify if the spindle frequency slope should be computed.
               "computeRMS"         : (optional) Specify if the spindle RMS amplitude should be computed.
               "computeFreq"        : (optional) Specify if the spindle average frequency  should be computed.
                          
     Outputs : "files"   : String indicating the full path to EEG file.
"""
class SpindleExtractionBWM(PyBlockWork.Block):

    def execute(self): 

        if self.verbose :        
            self.print_("Reading " + self.file + "...")
        
        # A reader class can be spedicied as en input but the EDFReader is used
        # by default.
        if hasattr(self, 'readerClass'):
            if not isinstance(self.readerClass, (str, unicode)) :
                raise TypeError("Expected type str or unicode, received type " + str(type(self.readerClass)) + ".")
            reader = apply(globals()[self.readerClass], [self.file])  
        else:
            reader =  EDFReader(self.file)              
                
            

        if not hasattr(self, 'eventName'):
            self.eventName = "SpindleRMS"
                    
        # Remove any events that may be preexisting in the files and which 
        # has the same name as the events we want to add to avoid any mixing up.
        reader.events.remove([e for e in reader.events if e.name == self.eventName])
        
        if self.verbose :
            self.print_("Detecting spindles in " + self.file + "...")
            
        # A reader class can be spedicied as en input but the EDFReader is used
        # by default.            
        if hasattr(self, 'detectorClass'):
            if not isinstance(self.detectorClass, (str, unicode)):
                raise TypeError("Expected type str or unicode, received type " + str(type(self.detectorClass)) + ".")
            keywords = {"reader":reader, "usePickled":False}  
            detector = apply(globals()[self.detectorClass], [], keywords)  
        else:
            detector   =  SpindleDectectorRMS(reader, usePickled=False)            
        

        if not hasattr(self, 'detectionStages'):
            self.detectionStages = ["Sleep stage 2"]      
        
        detector.setDetectionStages(self.detectionStages)
        
        if hasattr(self, 'detectionThreshold'):       
            detector.threshold = self.detectionThreshold
            
        if hasattr(self, 'computeSlopeFreq'):       
            detector.computeSlopeFreq = self.computeSlopeFreq

        if hasattr(self, 'computeRMS'):       
            detector.computeRMS = self.computeRMS

        if hasattr(self, 'computeFreq'):       
            detector.computeFreq = self.computeFreq


        detector.detectSpindles(listChannels=self.listChannels)

        if self.verbose :        
            self.print_("Saving spindles...") 
            
        detector.saveSpindle(reader, self.eventName, "Spindle")






    
"""
     Inputs  : "file" : String indicating the full path to a .sig file.
               "listChannels" : List of channel on which we want to extract 
                                the spindles(e.g., ['Fp1-ref', 'Fp2-ref', 'F7-ref')
               "resSubDir"    : Subdirectory where to save the results. If the directory
                                does not exist, it is created. (e.g., "Results/")
               "resFilesPrefix": Prefixe of the result files if results is to be
                                 stored in text files (e.g., "spindleDelays_offset").
               "dbEngine"     : Dabase engine if results are to be stored in a database.
               "offset"       : Time offset of the comparison in seconds (e.g., 5.0).
               "eventName"    : (optional) Specify the name of the spindle event.
                                 "SpindleRMS" is used by default.
                          
     Outputs : filesOut   : String indicating the full path to EEG file.
               pathOut    :
               listChannels :
               night      :
"""    
class SpindlePropagationBWM(PyBlockWork.Block):

    def execute(self):         
        
        if not hasattr(self, 'eventName'):
            self.eventName = "SpindleRMS"   

        
        session = Session()
        try:
            computeXCST(EDFReader, self.file, self.eventName, self.listChannels, dbSession=session,
                             delta =0.5, offset=self.offset, eventProperties = [])
            session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close()  






    
"""                
     Inputs : fileNameSyncLst   : String indicating the full path to EEG file.
              fileNameAsyncLst  : String indicating the full path to EEG file.
              path   :
              listChannels :
              night      :
              includeElectrodePatterns
              excludeElectrodePatterns
"""    
class ComputeSPF_BWM(PyBlockWork.Block):

    def execute(self):         
        
        # Keeping only the part of the channel name identifying the active electrode.
        electrodeList = map(lambda channel: getActiveElectrode(self.includeElectrodePatterns, 
                                                 self.excludeElectrodePatterns, channel), self.listChannels)        
        
        computeSPF(self.path, self.night, self.fileNameSyncList, 
                      self.fileNameAsyncLst, self.listChannels, electrodeList)        
                    
