# -*- coding: utf-8 -*-
"""
Created on Thu Oct 17 15:13:44 2013

@author: oreichri
"""

import socket

import sys, time, signal, traceback
from multiprocessing import Process
import Pyro4
from time import sleep
import os
from cStringIO import StringIO

from PyBlockWork import Block
from spyndle import terminate_process


###############################################################################
# Framework objects
###############################################################################



class ProcessingNode:
    def __init__(self, nbWorkers, pyroInfos={"stagingRoomName":"stagingRoom", 
                                             "comHubName":"communicationHub",
                                             "HMAC_KEY":""}):
                              
        if not isinstance(pyroInfos, dict):
            raise TypeError("pyroInfos must be a dict object.")
            
        for key in ["stagingRoomName", "comHubName", "HMAC_KEY"]:                                                 
            if not key in pyroInfos:
                raise NameError("pyrInfos must contain the key " + key + ".")
                                                 
        self.ip = socket.gethostbyname(socket.gethostname())
        self.workerPool = WorkerPool(nbWorkers, self.ip, pyroInfos)


    def terminate(self, force=False):
        self.workerPool.requestStop(force) 


class WorkerPool:
    def __init__(self, nbWorkers, ip, pyroInfos):
                                             
        self.pyroInfos = pyroInfos
        Pyro4.config.HMAC_KEY = pyroInfos["HMAC_KEY"]           
        
        sys.excepthook = Pyro4.util.excepthook    
        
        try:
            Pyro4.locateNS()
        except Pyro4.errors.PyroError, e:
            print 'Pyro error message: ' + str(e)
            raise
        except:
            raise        


        # Watch dog to stop the worker processes if the registered pyros objects
        #  becomes unavailable.Ensure that no orphan process stays alive
        # upon termination of mandatory pyro objects.
        self.watchdog = WhatchDog(pyroInfos["HMAC_KEY"])  
        self.watchdog.daemon = True  
        
        self.workers = {}
        for i in range(nbWorkers):
            workerID = ip + "-" + str(i)
            self.workers[workerID] = BlockWorker(workerID, pyroInfos, self.watchdog)
            self.workers[workerID].start()
            self.watchdog.registerProcess(self.workers[workerID].pid)

        
        self.watchdog.registerMonitoredObjects(pyroInfos["stagingRoomName"])
        self.watchdog.registerMonitoredObjects(pyroInfos["comHubName"])
        
        self.watchdog.start()                

        self.stopRequested = False


    def requestStop(self, force=False):
        self.stopRequested = True
        
        for workerID in self.workers:
            with Pyro4.Proxy("PYRONAME:" + self.pyroInfos["comHubName"]) as comHub:   
                comHub.terminateWorker(workerID)            
            
            self.workers[workerID].stopRequested = True
            if force:
                 os.kill(self.workers[workerID].pid, signal.SIGINT)
                 


"""
 Wrapping a process around the BWModule object because we don't want to be passing
 arround process objects. We want to be able to pass BWModule and to constitute
 this process object only when the BWModule run() method must be executed thought
 as start() call. 
"""
class BlockWorker(Process):

    def __init__(self, ID, pyroInfos, watchdog):
        
        super(BlockWorker, self).__init__()
        self.ID         = ID
        self.pyroInfos  = pyroInfos
        self.watchdog   = watchdog


        # Set the signal handler to catch process interuption process
        if sys.platform == 'win32':
            signal.signal(signal.SIGINT, self.exitProcess)
            signal.signal(signal.SIGTERM, self.exitProcess)
            signal.signal(signal.SIGABRT, self.exitProcess)
            
        # TODO: Add specific set of defined signals for the various platform
        else:
            signal.signal(signal.SIGINT, self.exitProcess)
            signal.signal(signal.SIGTERM, self.exitProcess)
            signal.signal(signal.SIGABRT, self.exitProcess)
            signal.signal(signal.SIGQUIT, self.exitProcess)
            signal.signal(signal.TSTP, self.exitProcess)
            signal.signal(signal.CTRL_C_EVENT, self.exitProcess)
            
        self.stopRequested = False
    
    
    def print_(self, string):
        try:
            with Pyro4.Proxy("PYRONAME:" + self.pyroInfos["comHubName"]) as comHub:   
                comHub.writeMessage(self.ID, string)  
        except Exception:
            print "Pyro traceback:"
            print "".join(Pyro4.util.getPyroTraceback())         
    
    
    def run(self): 
         

        try:
    
            Pyro4.config.HMAC_KEY = self.pyroInfos["HMAC_KEY"]           
            sys.excepthook = Pyro4.util.excepthook    
              
              
            with Pyro4.Proxy("PYRONAME:" + self.pyroInfos["comHubName"]) as comHub:   
                comHub.addWorker(self.ID )
                         
    
            while not self.stopRequested:
                
                with Pyro4.Proxy("PYRONAME:" + self.pyroInfos["stagingRoomName"]) as stagingRoom:   
                    blockString = stagingRoom.getReadyBlockString()

                
                self.BWMobj = Block.deserialize(blockString)
                if self.BWMobj is None:
                    sleep(1)
                    self.print_("No block in staging Room.")
                    
                else:
                    self.watchdog.registerMonitoredObjects(self.BWMobj.registerName)                
                    self.BWMobj.run(self.ID, self.pyroInfos)
                    self.watchdog.unregisterMonitoredObjects(self.BWMobj.registerName)                
                    
        except:
            tracebackOutput = StringIO()
            print "Trigger Exception, traceback info forward to log file."
            traceback.print_exc(file=tracebackOutput)
            self.print_(tracebackOutput.getvalue())
            sys.exit(1)



    def exitProcess(self):
        self.stopRequested = True
        self.BWMobj.print_("Exit signal has been catched.")
       
       
       
    @property
    def comHubName(self):
        return self.__comHubName
        
    @comHubName.setter
    def comHubName(self, comHubName = "communicationHub"):        
        self.__comHubName = comHubName

        
        
        

class WhatchDog(Process):
    
    def __init__(self, HMAC_KEY):
        self.clientPIDs        = []  
        self.monitoredObjects  = []
        Process.__init__(self)
        self.HMAC_KEY = HMAC_KEY
        
    def run(self):
        
        Pyro4.config.HMAC_KEY= self.HMAC_KEY
        while(True):
            time.sleep(5)
            try:
                for objName in self.monitoredObjects:
                    with Pyro4.Proxy("PYRONAME:" + objName) as obj:
                        obj.test()
                
            except Pyro4.errors.ConnectionClosedError:
                for PID in self.clientPIDs :
                    terminate_process(PID)
                break
            except Pyro4.errors.CommunicationError:
                for PID in self.clientPIDs :
                    terminate_process(PID)
                break


    def registerProcess(self, pid):
        if not pid in self.clientPIDs :
            self.clientPIDs.append(pid)

    def registerMonitoredObjects(self, objectName):
        if not objectName in self.monitoredObjects :
            self.monitoredObjects.append(objectName)
        
    def unregisterMonitoredObjects(self, objectName):
        if not objectName in self.monitoredObjects :
            self.monitoredObjects.remove(objectName)
        
        
        