# -*- coding: utf-8 -*-
"""
Created on Mon Oct 14 17:27:00 2013

@author: oreichri
"""

import serpent
import collections
import Pyro4

from PyBlockWork import StagingRoom, Block
from PyBlockWork.blocks import *


class Schema(collections.MutableMapping):
    

    def __init__(self, *args, **kwargs):
        self.store = dict()
        self.update(dict(*args, **kwargs))  # use the free update to set keys
        self.__startBlock = None
        self.__registerName = "register"
        
        self.__HMAC_KEY = ""

    @property
    def HMAC_KEY(self):
        return self.__HMAC_KEY

    @HMAC_KEY.setter
    def HMAC_KEY(self, value):
        if not isinstance(value, str):
            raise TypeError("HMAC_KEY property must be a str. A variable of type "\
                                                + type(value) + " has been passed.")
        self.__HMAC_KEY = value


    def __getitem__(self, key):
        return self.store[self.__keytransform__(key)]

    def __setitem__(self, key, value):
        self.store[self.__keytransform__(key)] = value

    def __delitem__(self, key):
        del self.store[self.__keytransform__(key)]
        
    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def __keytransform__(self, key):
        return key

    @property
    def startBlock(self):
        return self.__startBlock 

    @startBlock.setter
    def startBlock(self, name):
        if name in self:
            self.__startBlock = name
            self[name].registerName = self.__registerName
        else:
            raise ValueError("Cannot set " + name + " as the starting block because the schema has no such block.")
    
    def start(self, pyroInfos):
        
        
        if not "stagingRoomName" in pyroInfos:   
            raise ValueError("The 'stagingRoomName' key must be in the pyroInfos dict.")
        self.stagingRoomName = pyroInfos["stagingRoomName"]
            
        if self.startBlock is None:
            raise ValueError("Cannot start the schema because it has no starting block. First set one using the startBlock property.")            
          
        if "HMAC_KEY" in pyroInfos:
            self.HMAC_KEY = pyroInfos["HMAC_KEY"]
        
        for key in self.store:
            self.store[key].HMAC_KEY = self.HMAC_KEY
            
        if "registerName" in pyroInfos:
            self.registerName = pyroInfos["registerName"]
            for key in self.store:
                self.store[key].registerName = self.registerName
            

        for key in self.store:
            self.store[key].stagingRoomName = pyroInfos["stagingRoomName"]
            self.store[key].stagingRoomExec(StagingRoom.registerBlockString, self.store[key].ID, self.store[key].serialize())            
            
            
        self.stageBlock(self[self.startBlock].ID)


    def stageBlock(self, ID):
        try:
            with Pyro4.Proxy("PYRONAME:" + self.stagingRoomName) as stagingRoom:   
                stagingRoom.addReadyBlock(ID)
        except Exception:
            print "Pyro traceback:"
            print "".join(Pyro4.util.getPyroTraceback())       

    @property
    def registerName(self):
        return self.__registerName 
        
    @registerName.setter
    def registerName(self, registerName):
        
        if not self.startBlock is None:
            self[self.__startBlock].registerName = registerName        
        
        self.__registerName = registerName
        


    def save(self, fileName) :
        serpent.dump(self, open(fileName, "wb" ), indent=True, set_literals=False)
        
        
    @staticmethod
    def load(fileName) :

        serpentData = serpent.load(open(fileName, "rb"))
        
        if not serpentData["__class__"] == "Schema" :
            raise ValueError("The code contained in " + fileName + " is invalid.")
             
        schema = Schema()        
        
        schema.__startBlock = serpentData["_Schema__startBlock"]
        schema.__registerName = serpentData["_Schema__registerName"]
        schema.store = serpentData["store"]

        for key in schema.store:    
            schema.store[key] = Block.fromASTtoObject(schema.store[key])

        return schema
        
        